/*
 * Crackerjack Project
 *
 * Copyright (C) 2007-2008, Hitachi, Ltd.
 * Author(s): Takahiro Yasui <takahiro.yasui.mp@hitachi.com>,
 *            Yumiko Sugita <yumiko.sugita.yf@hitachi.com>,
 *            Satoshi Fujiwara <sa-fuji@sdl.hitachi.co.jp>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * $Id:$
 *
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include <asm/errno.h>
#include "../../include/socketcall.h"


/*
 * Macros
 */
#define SYSCALL_NAME	"getpeername"


/*
 * Global variables
 */
static int opt_debug;
static char *progname;

enum test_type {
	FD_NONE,
	FD_NOT_EXIST,
	FD_FILE,
	NOT_CONN,
	NULL_ADDR,
	NULL_LEN,
	NORMAL,
};


/*
 * Data Structure
 */
struct test_case {
	int domain;
	int ttype;
	int ret;
	int err;
};


/* Test cases
 *
 *   test status of errors on man page
 *
 *   EBADF		v (not a valid descriptor)
 *   ENOTSOCK		v (file descriptor)
 *   ENOTCONN		v (not connected)
 *   ENOBUFS		can't check because it's difficult to create no-memory
 *   EFAULT		v (outside user's space)
 */
static struct test_case tcase[] = {
	// -------- PF_UNIX --------
	{ // case00
		.domain	= PF_UNIX,
		.ttype	= FD_NONE,
		.ret	= -1,
		.err	= EBADF,
	},
	{ // case01
		.domain	= PF_UNIX,
		.ttype	= FD_NOT_EXIST,
		.ret	= -1,
		.err	= EBADF,
	},
	{ // case02
		.domain	= PF_UNIX,
		.ttype	= FD_FILE,
		.ret	= -1,
		.err	= ENOTSOCK,
	},
	{ // case03
		.domain	= PF_UNIX,
		.ttype	= NULL_ADDR,
		.ret	= -1,
		.err	= EFAULT,
	},
	{ // case04
		.domain	= PF_UNIX,
		.ttype	= NULL_LEN,
		.ret	= -1,
		.err	= EFAULT,
	},
	{ // case05
		.domain	= PF_UNIX,
		.ttype	= NOT_CONN,
		.ret	= -1,
		.err	= ENOTCONN,
	},
	{ // case06
		.domain	= PF_UNIX,
		.ttype	= NORMAL,
		.ret	= 0,
		.err	= 0,
	},
	// -------- PF_INET --------
	{ // case07
		.domain	= PF_INET,
		.ttype	= FD_NONE,
		.ret	= -1,
		.err	= EBADF,
	},
	{ // case08
		.domain	= PF_INET,
		.ttype	= FD_NOT_EXIST,
		.ret	= -1,
		.err	= EBADF,
	},
	{ // case09
		.domain	= PF_INET,
		.ttype	= FD_FILE,
		.ret	= -1,
		.err	= ENOTSOCK,
	},
	{ // case10
		.domain	= PF_INET,
		.ttype	= NULL_ADDR,
		.ret	= -1,
		.err	= EFAULT,
	},
	{ // case11
		.domain	= PF_INET,
		.ttype	= NULL_LEN,
		.ret	= -1,
		.err	= EFAULT,
	},
	{ // case12
		.domain	= PF_INET,
		.ttype	= NOT_CONN,
		.ret	= -1,
		.err	= ENOTCONN,
	},
	{ // case13
		.domain	= PF_INET,
		.ttype	= NORMAL,
		.ret	= 0,
		.err	= 0,
	},
};

/*
 * do_test()
 *
 *   Input  : TestCase Data
 *   Return : RESULT_OK(0), RESULT_NG(1)
 *
 */
static int do_test(struct test_case *tc)
{
	int fd = -1, rc;
	int sys_ret;
	int sys_errno;
	int result = RESULT_OK;
	struct sockaddr_in in_addr, rin_addr;
	struct sockaddr_un un_addr, run_addr;
	struct sockaddr *addr, *raddr;
	int len, rlen;
	pid_t sv_pid = 0;
	int cmp_ok = 1;
	int st;

	// initialize 'xn_addr' and 'len' and create server process if necessary
	if (tc->domain == PF_UNIX) {
		un_addr.sun_family = AF_UNIX;
		strcpy(un_addr.sun_path, "");
		len = sizeof(un_addr);
		addr = (struct sockaddr*)&un_addr;
		raddr = (struct sockaddr*)&run_addr;
	} else {
		in_addr.sin_family = AF_INET;
		in_addr.sin_port = 11000;	// this port must be unused
		in_addr.sin_addr.s_addr = INADDR_ANY;
		len = sizeof(in_addr);
		addr = (struct sockaddr*)&in_addr;
		raddr = (struct sockaddr*)&rin_addr;
	}
	sv_pid = create_server_proc(SOCK_STREAM, 0, addr, len, SV_NORMAL);
	if (sv_pid < 0) {
		result = 1;
		goto EXIT;
	}

	switch (tc->ttype) {
	case FD_NOT_EXIST:
		fd = INT_MAX - 1;
		/* fallthrough */
	case FD_NONE:
		break;
	case NULL_ADDR:
	case NULL_LEN:
	case NORMAL:
	case NOT_CONN:
		fd = socket(tc->domain, SOCK_STREAM, 0);
		if (fd < 0) {
			EPRINTF("can't create socket.\n");
			result = 1;
			goto EXIT;
		}
		break;
	case FD_FILE:
		fd = open("/", O_RDONLY);
		if (fd < 0) {
			EPRINTF("can't open \"/\".\n");
			return 1;
		}
		break;
	}
	// do connect
	switch (tc->ttype) {
	case NULL_ADDR:
	case NULL_LEN:
	case NORMAL:
		rc = connect(fd, addr, len);
		if (rc < 0) {
			EPRINTF("connect failed.\n");
			result = 1;
			goto EXIT;
		}
		break;
	}
	// do getpeername
	errno = 0;
	switch (tc->ttype) {
	case NULL_ADDR:
		sys_ret = getpeername(fd, NULL, &rlen);
		break;
	case NULL_LEN:
		sys_ret = getpeername(fd, raddr, NULL);
		break;
	default:
		sys_ret = getpeername(fd, raddr, &rlen);
		break;
	}
	sys_errno = errno;
	if (sys_ret >= 0) {
		if (tc->domain == PF_UNIX)
			cmp_ok = un_addr.sun_family == run_addr.sun_family &&
				strcmp(un_addr.sun_path, run_addr.sun_path) ==0;
		else
			cmp_ok = in_addr.sin_family == rin_addr.sin_family &&
				in_addr.sin_port == rin_addr.sin_port;
				//in_addr.sin_addr.s_addr ==
				//	rin_addr.sin_addr.s_addr;
	}

	/*
	 * Check results
	 */
	result |= (sys_errno != tc->err) || !cmp_ok;
	PRINT_RESULT_CMP(sys_ret >= 0, tc->ret, tc->err, sys_ret, sys_errno,
			 cmp_ok);
EXIT:
	if (fd >= 0)
		close(fd);
	if (sv_pid > 0) {
		kill(sv_pid, SIGTERM);
		wait(&st);
	}
	return result;
}


/*
 * usage()
 */
static void usage(const char *progname)
{
	EPRINTF("usage: %s [options]\n", progname);
	EPRINTF("This is a regression test program of %s system call.\n",
		SYSCALL_NAME);
	EPRINTF("options:\n");
	EPRINTF("    -d --debug           Show debug messages\n");
	EPRINTF("    -h --help            Show this message\n");

	RPRINTF("NG\n");
	exit(1);
}


/*
 * main()
 */
int main(int argc, char *argv[])
{
	int result = RESULT_OK;
	int c;
	int i;

	struct option long_options[] = {
		{ "debug", no_argument, 0, 'd' },
		{ "help",  no_argument, 0, 'h' },
		{ NULL, 0, NULL, 0 }
	};

	progname = strchr(argv[0], '/');
	progname = progname ? progname + 1 : argv[0];

	while ((c = getopt_long(argc, argv, "dh", long_options, NULL)) != -1) {
		switch (c) {
		case 'd':
			opt_debug = 1;
			break;

		default:
			usage(progname);
			// NOTREACHED
		}
	}

	if (argc != optind) {
		EPRINTF("Options are not match.\n");
		usage(progname);
		// NOTREACHED
	}

	/*
	 * Execute test
	 */
	for (i = 0; i < (int)(sizeof(tcase) / sizeof(tcase[0])); i++) {
		int ret;

		PRINTF("(case%02d) START\n", i);
		ret = do_test(&tcase[i]);
		PRINTF("(case%02d) END => %s\n", i, (ret == 0) ? "OK" : "NG");

		result |= ret;
	}

	/*
	 * Check results
	 */
	switch(result) {
	case RESULT_OK:
		RPRINTF("OK\n");
		break;

	default:
		RPRINTF("NG\n");
		break;
	}

	return 0;
}

/*
 * gcc -W -Wall -O2 -o test <filename>.c
 */
